#!/usr/bin/env python
# -*- encoding: latin-1 -*-
# -*- coding: latin-1 -*-
# Modulo que incluye las clases para implementar un cliente y 
# servidor xmlrpc con soporte para compresion de datos.
#
from SimpleXMLRPCServer import SimpleXMLRPCServer, SimpleXMLRPCRequestHandler
import zlib
from xmlrpclib import Transport, ProtocolError

#Constantes
#Define si se usara compresion gzip (cliente)
GZIP_ENCODED = True


class ICTCXmlRpcServer(SimpleXMLRPCServer):
	'''
	Servidor XMLRPC que soporta comunicaciones gzipeadas.
	'''
	_version = '1.00.00'
	_author = 'Marco Antonio Islas Cruz <mislas@ictc.com.mx>'
	def __init__(self, *args, **kargs):
		# Usamos una clase interna, asi podemos llamar al metodo post.
		class ICTCRequestHandler(SimpleXMLRPCRequestHandler):
			#Inicializamos la clase padre asi como la bandera gzipped
			def __init__(self, *args,**kwargs):
				SimpleXMLRPCRequestHandler.__init__(self, *args, **kwargs)
				self.gzipped = False
			
			#Este metodo maneja los requests.
			def parse_request(request):
				# Primero, llamamos la implementacion original que nos regresa
				# True si todo va bien
				if SimpleXMLRPCRequestHandler.parse_request(request):
					# Buscamos por el encabezado de compresion 
					# y que este sea "gzip"
					self.gzipped = 'accept-encoding' in request.headers.keys() \
							and request.headers['accept-encoding'] == 'gzip'
					return True
				else:
					# Si el metodo falla entonces notificar al cliente
					request.send_error(500, 'Internal Server Error')
				return False

			def do_POST(self):
				if self.headers.has_key('accept-encoding') and \
						self.headers['accept-encoding'] == 'gzip':
					# Revisar si la ruta es correcta
					if not self.is_rpc_path_valid():
						self.report_404()
						return
					try: #Intentar obtener los datos
						max_chunk_size = 10*1024*1024
						size_remaining = int(self.headers["content-length"])
						L = []
						while size_remaining>0:
							chunk_size = min(size_remaining, max_chunk_size)
							L.append(self.rfile.read(chunk_size))
							size_remaining -= len(L[-1])
						data = ''.join(L)
						data = zlib.decompress(data) #Descomprimir los datos
						response = self.server._marshaled_dispatch(
								data, getattr(self, '_dispatch', None)
								)
					except:
						#Seria bueno registrar en algun lado la excepcion
						pass
					else:
						# Tenemos una respuesta XML RPC valida
						#comprimir los datos
						response = zlib.compress(response)
						self.send_response(200)
						self.send_header("Content-type", "text/xml")
						self.send_header("Content-length", str(len(response)))
						# Como la pregunta fue comprimida la respuesta va 
						# comprimida tambien
						self.send_header("Accept-Encoding","gzip")
						self.end_headers()
						self.wfile.write(response)
						# Cerrar la conexion
						self.wfile.flush()
						self.connection.shutdown(1)
				else:
					return SimpleXMLRPCRequestHandler.do_POST(self)
		# Inicializar la superclase
		SimpleXMLRPCServer.__init__(self,requestHandler=ICTCRequestHandler,
									*args, **kargs)




# ----------------------- Transport -------------------------
# Esta clase permite que un cliente xmlrpc pueda enviar y recibir datos
# comprimidos de un servidor xmlrpc

class ICTCTransport(object, Transport):
	'''
	Esta clase debe ser usada como Transport (HTTPTransport) para un cliente
	XMLRPC ej.
	xmlrpclib.ServerProxy("http://localhost:8000/", transport = trans_inst)
	'''
	_version = '1.00.00'
	_author = 'Marco Antonio Islas Cruz <mislas@ictc.com.mx>'
	def __init__(self, *args,**kwargs):
		Transport.__init__(self, *args, **kwargs)
		self.__gzip_encoded = False
	
	def request(self, host, handler, request_body, verbose=0):
		'''
		Encargado de manejar los request, envia los datos y recibe
		las repuestas, por lo que aqui se hace la compresion de datos
		y posteriormente se envian los datos recibidos a 
		la funcion self._parse_response o self._parse_gzipped_response
		de acuerdo al caso.
		Mas informacion:
		http://docs.python.org/library/xmlrpclib.html
		'''
		h = self.make_connection(host)
		if verbose:
			h.set_debuglevel(1)
		self.send_request(h, handler, request_body)
		self.send_host(h, host)
		self.send_user_agent(h)
		# Comprimir los datos si GZIP_ENCODED es True
		if GZIP_ENCODED: 
			# No olvidar de poner el encabezado de compresion
			h.putheader('Accept-Encoding','gzip')
			request_body = zlib.compress(request_body)
		self.send_content(h, request_body)
		
		errcode, errmsg, headers = h.getreply()
		if errcode != 200:
			raise ProtocolError(
					host + handler,
					errcode, errmsg,
					headers
					)
		self.verbose = verbose
		try:
			sock = h._conn.sock
		except AttributeError:
			sock = None
		# Revisar si la respuesta incluye los encabezados de 
		# compresion y si este es gzip
		if headers.has_key('accept-encoding') and \
				headers['accept-encoding'] == 'gzip':
			return self._parse_gzipped_response(h.getfile(), sock)
		return self._parse_response(h.getfile(), sock)

	def _parse_gzipped_response(self, file, sock):
		'''
		Practicamente lo mismo que _parse_response solo que 
		agregamos el uso de zlib.decompress
		'''
		p, u = self.getparser()
		response = ''
		while 1:
			if sock:
				r = sock.recv(1024)
			else:
				r = file.read(1024)
			if not r:
				break
			response += r
		response = zlib.decompress(response)
		if self.verbose:
			print "body:", repr(response)
		p.feed(response)
		file.close()
		p.close()
		return u.close()
